/*
	cmd.c
*/

#include <stdio.h>

#include "cmd.h"
#include "types.h"

vm_command_t test_commands[] = {
	{0x00, "",			0},
	{0x01, "equaln",		2},
	{0x02, "equalv",		2},
	{0x03, "lessn",			2},
	{0x04, "lessv",			2},
	{0x05, "greatern",		2},
	{0x06, "greaterv",		2},
	{0x07, "isset",			1},
	{0x08, "issetv",		1},
	{0x09, "has",			1},
	{0x0A, "obj.in.room",		2},
	{0x0B, "posn",			5},
	{0x0C, "controller",		1},
	{0x0D, "have.key",		0},
	{0x0E, "said",			0},
	{0x0F, "compare.strings",	2},
	{0x10, "obj.in.box",		5},
	{0x11, "center.posn",		5},
	{0x12, "right.posn",		5},
};

int total_test_commands =
	(sizeof (test_commands) / sizeof (test_commands[0]));

vm_command_t action_commands[] = {
	{0x00, "return", 		0},
	{0x01, "increment", 		1},
	{0x02, "decrement", 		1},
	{0x03, "assignn", 		2},
	{0x04, "assignv", 		2},
	{0x05, "addn", 			2},
	{0x06, "addv", 			2},
	{0x07, "subn", 			2},
	{0x08, "subv", 			2},
	{0x09, "lindirectv",		2},
	{0x0A, "rindirect",		2},
	{0x0B, "lindirectn",		2},
	{0x0C, "set",			1},
	{0x0D, "reset",			1},
	{0x0E, "toggle",		1},
	{0x0F, "set.v",			1},
	{0x10, "reset.v",		1},
	{0x11, "toggle.v",		1},
	{0x12, "new.room",		1},
	{0x13, "new.room.v",		1},
	{0x14, "load.logic",		1},
	{0x15, "load.logic.v",		1},
	{0x16, "call",			1},
	{0x17, "call.v",		1},
	{0x18, "load.pic",		1},
	{0x19, "draw.pic",		1},
	{0x1A, "show.pic",		0},
	{0x1B, "discard.pic",		1},
	{0x1C, "overlay.pic",		1},
	{0x1D, "show.pri.screen",	0},
	{0x1E, "load.view",		1},
	{0x1F, "load.view.v",		1},
	{0x20, "discard.view",		1},
	{0x21, "animate.obj",		1},
	{0x22, "unanimate.all",		0},
	{0x23, "draw",			1},
	{0x24, "erase",			1},
	{0x25, "position",		3},
	{0x26, "position.v",		3},
	{0x27, "get.posn",		3},
	{0x28, "reposition",		3},
	{0x29, "set.view",		2},
	{0x2A, "set.view.v",		2},
	{0x2B, "set.loop",		2},
	{0x2C, "set.loop.v",		2},
	{0x2D, "fix.loop",		1},
	{0x2E, "release.loop",		1},
	{0x2F, "set.cel",		2},
	{0x30, "set.cel.v",		2},
	{0x31, "last.cel",		2},
	{0x32, "current.cel",		2},
	{0x33, "current.loop",		2},
	{0x34, "current.view",		2},
	{0x35, "number.of.loops",	2},
	{0x36, "set.priority",		2},
	{0x37, "set.priority.v",	2},
	{0x38, "release.priority",	1},
	{0x39, "get.priority",		2},
	{0x3A, "stop.update",		1},
	{0x3B, "start.update",		1},
	{0x3C, "force.update",		1},
	{0x3D, "ignore.horizon",	1},
	{0x3E, "observe.horizon",	1},
	{0x3F, "set.horizon",		1},
	{0x40, "object.on.water",	1},
	{0x41, "object.on.land",	1},
	{0x42, "object.on.anything",	1},
	{0x43, "ignore.objs",		1},
	{0x44, "observe.objs",		1},
	{0x45, "distance",		3},
	{0x46, "stop.cycling",		1},
	{0x47, "start.cycling",		1},
	{0x48, "normal.cycle",		1},
	{0x49, "end.of.loop",		2},
	{0x4A, "reverse.cycle",		1},
	{0x4B, "reverse.loop",		2},
	{0x4C, "cycle.time",		2},
	{0x4D, "stop.motion",		1},
	{0x4E, "start.motion",		1},
	{0x4F, "step.size",		2},
	{0x50, "step.time",		2},
	{0x51, "move.obj",		5},
	{0x52, "move.obj.v",		5},
	{0x53, "follow.ego",		3},
	{0x54, "wander",		1},
	{0x55, "normal.motion",		1},
	{0x56, "set.dir",		2},
	{0x57, "get.dir",		2},
	{0x58, "ignore.blocks",		1},
	{0x59, "observe.blocks",	1},
	{0x5A, "block",			4},
	{0x5B, "unblock",		0},
	{0x5C, "get",			1},
	{0x5D, "get.v",			1},
	{0x5E, "drop",			1},
	{0x5F, "put",			2},
	{0x60, "put.v",			2},
	{0x61, "get.room.v",		2},
	{0x62, "load.sound",		1},
	{0x63, "sound",			2},
	{0x64, "stop.sound",		0},
	{0x65, "print",			1},
	{0x66, "print.v",		1},
	{0x67, "display",		3},
	{0x68, "display.v",		3},
	{0x69, "clear.lines",		3},
	{0x6A, "text.screen",		0},
	{0x6B, "graphics",		0},
	{0x6C, "set.cursor.char",	1},
	{0x6D, "set.text.attribute",	2},
	{0x6E, "shake.screen",		1},
	{0x6F, "configure.screen",	3},
	{0x70, "status.line.on",	0},
	{0x71, "status.line.off",	0},
	{0x72, "set.string",		2},
	{0x73, "get.string",		5},
	{0x74, "word.to.string",	2},
	{0x75, "parse",			1},
	{0x76, "get.num",		2},
	{0x77, "prevent.input",		0},
	{0x78, "accept.input",		0},
	{0x79, "set.key",		3},
	{0x7A, "add.to.pic",		7},
	{0x7B, "add.to.pic.v",		7},
	{0x7C, "status",		0},
	{0x7D, "save.game",		0},
	{0x7E, "restore.game",		0},
	{0x7F, "init.disk",		0},
	{0x80, "restart.game",		0},
	{0x81, "show.obj",		1},
	{0x82, "random",		3},
	{0x83, "program.control",	0},
	{0x84, "player.control",	0},
	{0x85, "obj.status.v",		1},
	{0x86, "quit",			1},
	{0x87, "show.mem",		0},
	{0x88, "pause",			0},
	{0x89, "echo.line",		0},
	{0x8A, "cancel.line",		0},
	{0x8B, "init.joy",		0},
	{0x8C, "toggle.monitor",	0},
	{0x8D, "version",		0},
	{0x8E, "script.size",		1},
	{0x8F, "set.game.id",		1},
	{0x90, "log",			1},
	{0x91, "set.scan.start",	0},
	{0x92, "reset.scan.start",	0},
	{0x93, "reposition.to",		3},
	{0x94, "reposition.to.v",	3},
	{0x95, "trace.on",		0},
	{0x96, "trace.info",		3},
	{0x97, "print.at",		4},
	{0x98, "print.at.v",		4},
	{0x99, "discard.view.v",	1},
	{0x9A, "clear.text.rect",	5},
	{0x9B, "set.upper.left",	2},
	{0x9C, "set.menu",		1},
	{0x9D, "set.menu.item",		2},
	{0x9E, "submit.menu",		0},
	{0x9F, "enable.item",		1},
	{0xA0, "disable.item",		1},
	{0xA1, "menu.input",		0},
	{0xA2, "show.obj.v",		1},
	{0xA3, "open.dialogue",		0},
	{0xA4, "close.dialogue",	0},
	{0xA5, "mul.n",			2},
	{0xA6, "mul.v",			2},
	{0xA7, "div.n",			2},
	{0xA8, "div.v",			2},
	{0xA9, "close.window",		0},
	{0xAA, "unknown170",		1},
	{0xAB, "unknown171",		0},
	{0xAC, "unknown172",		0},
	{0xAD, "unknown173",		0},
	{0xAE, "unknown174",		1},
	{0xAF, "unknown175",		1},
	{0xB0, "unknown176",		0},
	{0xB1, "unknown177",		1},
	{0xB2, "unknown178",		0},
	{0xB3, "unknown179",		4},
	{0xB4, "unknown180",		2},
	{0xB5, "unknown181",		0},
};

int total_action_commands =
	(sizeof (action_commands) / sizeof (action_commands[0]));
