/*
	save.c
*/

#include <stdio.h>
#include <string.h>

#include "graphics.h"
#include "lagii.h"
#include "logic.h"
#include "object.h"
#include "picture.h"
#include "resource.h"
#include "save.h"
#include "text.h"
#include "types.h"


void save_save (void)
{
	int err;

	err = save_save_game ("save0.gasp", "");
	if (err == -1) {
		message_box ("Error saving to 'save0.gasp'.", true);
	}
}

void save_restore (void)
{
	int err;

	err = save_restore_game ("save0.gasp");
	if (err == -1) {
		message_box ("Error restoring from 'save0.gasp'.", true);
	}
}

int save_save_game (char *fname, char *desc)
{
	FILE *f;
	int i, j;
	char *gid;
	l_byte ch, settings[16], flags[32];
	gasp_real_view_t view;
	agi_object_t *ob;

	f = fopen (fname, "w");
	if (!f)
		return -1;

	/*=================*
	 * (1) GASP header *
	 *=================*/
	fwrite ("GASP", 1, 4, f);

	gid = sierra_game[game_id].short_title;
	j = strlen (gid);
	for (i=0; i < j; i++) {
		ch = (l_byte) gid[i];
		if ((ch >= 'a') && (ch <= 'z'))
			ch -= ('a' - 'A');
		fputc (ch, f);
	}
	for (i=j; i < 8; i++)
		fputc (0, f);

	if (desc)
		ch = strlen (desc);
	else
		ch = 0;
	fputc (ch, f);
	if (ch > 0)
		fwrite (desc, 1, ch, f);

	ch = (last_picture & 0xFF);
	fputc (ch, f);
	memset (settings, 0, 16);
	settings[0] = (agi_horizon & 0xFF);
	settings[1] = (foreground_color & 0xFF);
	settings[2] = (background_color & 0xFF);
	settings[3] = (min_line & 0xFF);
	settings[4] = (user_input_line & 0xFF);
	settings[5] = (status_line & 0xFF);
	settings[6] = input_cursor;
	fwrite (settings, 1, 16, f);

	/*==========================*
	 * (2) Variable/Flag status *
	 *==========================*/
	fwrite (vm_variables, 1, 256, f);
	memset (flags, 0, 32);
	for (i=0; i < 256; i++) {
		if (vm_flags[i] == true)
			flags[(i >> 3)] |= (1 << (i & 0x07));
	}
	fwrite (flags, 1, 32, f);

	/*===================*
	 * (3) String status *
	 *===================*/
	for (i=0; i < 24; i++) {
		fputs (vm_strings[i], f);
		fputc (0, f);
	}

	/*=================*
	 * (4) View status *
	 *=================*/
	for (i=0; i < 256; i++) {
		ob = &agi_objects[i];

		view.x = (ob->x & 0xFFFF);
		view.y = (ob->y & 0xFFFF);

		view.view = ob->current_view;
		view.loop = ob->current_loop;
		view.cel = ob->current_cel;
		view.priority = ob->priority;

		view.step_size = ob->step_size;
		view.step_time = ob->step_time;
		view.cycle_time = ob->cycle_time;

		view.tx = ob->tx;
		view.ty = ob->ty;
		view.target_flag = ob->target_flag;

		view.flags = ob->flags;

		fwrite (&view, 1, sizeof (gasp_real_view_t), f);
	}

	/*======================*
	 * (5) Inventory status *
	 *======================*/
	for (i=0; i < 256; i++)
		fputc (agi_inventory_objects[i].room, f);

	fclose (f);
{
	char buf[50];
	sprintf (buf, "Saved game as '%s'.", fname);
	message_box (buf, true);
}
	return 0;
}

int save_restore_game (char *fname)
{
	FILE *f;
	int i, j;
	char *gid, msg[30];
	l_byte buf[32], ch, settings[16], flags[32];
	gasp_real_view_t view;
	agi_object_t *ob;

	f = fopen (fname, "r");
	if (!f)
		return -1;

	/*=================*
	 * (1) GASP header *
	 *=================*/
	fread (buf, 1, 4, f);
	if (strncmp ((char *) buf, "GASP", 4)) {
		fclose (f);
		message_box ("Not a GASP save game.", true);
		return -2;
	}
	gid = sierra_game[game_id].short_title;
	j = strlen (gid);
	memset (buf, 0, 9);
	fread (buf, 1, 8, f);
	for (i=0; i < j; i++) {
		ch = buf[i];
		if ((ch >= 'A') && (ch <= 'Z'))
			ch += ('a' - 'A');
		if (ch != ((l_byte) gid[i]))
			break;
	}
	if (i < j) {
		fclose (f);
		sprintf (msg, "Wrong game id (%s).", buf);
		message_box (msg, true);
		return -2;
	}
	ch = fgetc (f);
	if (ch > 0)
		fread (buf, 1, ch, f);
	buf[ch] = 0;
	/* (Description is NULL-terminated in buf) */
	ch = fgetc (f);
	draw_picture (PICTURE[ch], true);
	fread (settings, 1, 16, f);
	agi_horizon = settings[0];
	foreground_color = settings[1];
	background_color = settings[2];
	min_line = settings[3];
	user_input_line = settings[4];
	status_line = settings[5];
	input_cursor = settings[6];

	/*==========================*
	 * (2) Variable/Flag status *
	 *==========================*/
	fread (vm_variables, 1, 256, f);
	fread (flags, 1, 32, f);

	for (i=0; i < 256; i++) {
		j = (flags[(i >> 3)] & (1 << (i & 0x07)));
		if (j)
			vm_flags[i] = true;
		else
			vm_flags[i] = false;
	}

	/*===================*
	 * (3) String status *
	 *===================*/
	for (i=0; i < 24; i++) {
		memset (vm_strings[i], 0, 40);
		j = 0;
		while (1) {
			ch = fgetc (f);
			if (ch < 1)
				break;
			vm_strings[i][j++] = ch;
		}
	}

	/*=================*
	 * (4) View status *
	 *=================*/
	for (i=0; i < 256; i++) {
		fread (&view, 1, sizeof (gasp_real_view_t), f);
		ob = &agi_objects[i];

		ob->x = view.x;
		ob->y = view.y;

		ob->current_view = view.view;
		ob->current_loop = view.loop;
		ob->current_cel = view.cel;
		ob->priority = view.priority;

		ob->step_size = view.step_size;
		ob->step_time = view.step_time;
		ob->cycle_time = view.cycle_time;

		ob->tx = view.tx;
		ob->ty = view.ty;
		ob->target_flag = view.target_flag;

		ob->flags = view.flags;
	}

	/*======================*
	 * (5) Inventory status *
	 *======================*/
	for (i=0; i < 256; i++)
		agi_inventory_objects[i].room = fgetc (f);

{
	char msg[100];
	sprintf (msg, "Restored game from '%s'.\n\"%s\"", fname, buf);
	message_box (msg, true);
}
	fclose (f);
	return 0;
}
